# Copyright 2023 OmniSafe Team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import datetime
import os

import numpy as np


color2num = {
    'gray': 30,
    'red': 31,
    'green': 32,
    'yellow': 33,
    'blue': 34,
    'magenta': 35,
    'cyan': 36,
    'white': 37,
    'crimson': 38,
}


def colorize(string, color, bold=False, highlight=False):
    """
    Colorize a string.

    This function was originally written by John Schulman.
    """
    attr = []
    num = color2num[color]
    if highlight:
        num += 10
    attr.append(str(num))
    if bold:
        attr.append('1')
    return '\x1b[{}m{}\x1b[0m'.format(';'.join(attr), string)


class Logger:
    def __init__(self, algo, task, seed) -> None:
        self.curr_time = f'{datetime.datetime.now():%Y-%m-%d-%H:%M:%S}'
        self.first_row = True
        self.log_headers = []
        self.log_current_row = {}
        self.output_dir = './data'
        self.output_dir = os.path.join(self.output_dir)
        if os.path.exists(self.output_dir):
            print(
                'Warning: Log dir %s already exists! Storing info there anyway.' % self.output_dir,
            )
        else:
            os.makedirs(self.output_dir)
        st = '_'.join([self.curr_time, algo, task, str(seed), 'progress.txt'])
        self.output_file = open(os.path.join(self.output_dir, st), 'w')
        atexit.register(self.output_file.close)

    def log(self, msg, color='green'):
        """Print a colorized message to stdout."""
        print(colorize(msg, color, bold=True))

    def log_tabular(self, key, val):
        """
        Log a value of some diagnostic.

        Call this only once for each diagnostic quantity, each iteration.
        After using ``log_tabular`` to store values for each diagnostic,
        make sure to call ``dump_tabular`` to write them out to file and
        stdout (otherwise they will not get saved anywhere).
        """
        if self.first_row:
            self.log_headers.append(key)
        else:
            assert key in self.log_headers, (
                "Trying to introduce a new key %s that you didn't include in the first "
                'iteration' % key
            )
        assert key not in self.log_current_row, (
            'You already set %s this iteration. Maybe you forgot to call dump_tabular()' % key
        )
        self.log_current_row[key] = val

    def dump_tabular(self):
        """
        Write all of the diagnostics from the current iteration.

        Writes both to stdout, and to the output file.
        """
        vals = []
        key_lens = [len(key) for key in self.log_headers]
        max_key_len = max(15, max(key_lens))
        keystr = '%' + '%d' % max_key_len
        fmt = '| ' + keystr + 's | %15s |'
        n_slashes = 22 + max_key_len
        print('-' * n_slashes)
        for key in self.log_headers:
            val = self.log_current_row.get(key, '')
            valstr = '%8.3g' % val if hasattr(val, '__float__') else val
            print(fmt % (key, valstr))
            vals.append(val)
        print('-' * n_slashes, flush=True)
        if self.output_file is not None:
            if self.first_row:
                self.output_file.write('\t'.join(self.log_headers) + '\n')
            self.output_file.write('\t'.join(map(str, vals)) + '\n')
            self.output_file.flush()

        self.log_current_row.clear()
        self.first_row = False


class EpochLogger(Logger):
    def __init__(self, algo, task, seed) -> None:
        super().__init__(algo, task, seed)
        self.epoch = 0
        self.epoch_dict = {}

    def setup_global_epoch(self, epoch=None):
        self.epoch = epoch

    def store(self, **kwargs):
        """
        Save something into the epoch_logger's current state.

        Provide an arbitrary number of keyword arguments with numerical values.
        """
        for k, v in kwargs.items():
            if k not in self.epoch_dict.keys():
                self.epoch_dict[k] = []
            self.epoch_dict[k].append(v)

    def log_tabular(
        self,
        key,
        val=None,
        with_min_and_max=False,
        average_only=False,
        to_tb=False,
        to_sacred=False,
    ):
        if val is not None:
            super().log_tabular(key, val)
        else:
            v = self.epoch_dict[key]
            vals = np.concatenate(v) if isinstance(v[0], np.ndarray) and len(v[0].shape) > 0 else v
            stats = (np.mean(vals), np.std(vals), np.min(vals), np.max(vals))
            super().log_tabular(key if average_only else 'Average' + key, stats[0])
            if not (average_only):
                super().log_tabular('Std' + key, stats[1])
            if with_min_and_max:
                super().log_tabular('Max' + key, stats[3])
                super().log_tabular('Min' + key, stats[2])
        self.epoch_dict[key] = []
